@@ -1,5 +1,5 @@
 use crate::replica::ReplicaID;
-use crate::state_machine::StateMachineTransition;
+use crate::state_machine::{StateMachineTransition};
 
 /// LogEntry is a state machine transition along with some metadata needed for
 /// Raft.
@@ -16,14 +16,15 @@ where
 /// Message describes messages that the replicas pass between each other to
 /// achieve consensus on the distributed state machine.
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd)]
-pub enum Message<T>
+pub enum Message<T, D>
 where
     T: StateMachineTransition,
+    D: Clone,
 {
     /// AppendEntryRequest is used by the Leader to send out logs for other
     /// replicas to append to their log. It also has information on what logs
     /// are ready to be applied to the state machine. AppendEntryRequest is also
-    /// used as a heart beat message by the Leader even when no new logs need to
+    /// used as a heartbeat message by the Leader even when no new logs need to
     /// be processed.
     AppendEntryRequest {
         from_id: ReplicaID,
@@ -58,4 +59,20 @@ where
         term: usize,
         vote_granted: bool,
     },
+
+    InstallSnapshotRequest {
+        from_id: ReplicaID,
+        term: usize,
+        last_included_index: usize,
+        last_included_term: usize,
+        offset: usize,
+        data: D,
+        done: bool,
+    },
+
+    InstallSnapshotResponse {
+        from_id: ReplicaID,
+        term: usize,
+        last_included_index: usize,
+    },
 }
