@@ -1,11 +1,14 @@
 use crate::{
     cluster::Cluster,
     message::{LogEntry, Message},
-    state_machine::{StateMachine, StateMachineTransition, TransitionState},
+    state_machine::{
+        Snapshot, StateMachine, StateMachineTransition, TransitionAbandonedReason, TransitionState,
+    },
     timer::Timer,
 };
 use crossbeam_channel::{Receiver, Select};
 use rand::Rng;
+use std::cmp::Ordering;
 use std::sync::{Arc, Mutex};
 use std::{
     cmp,
@@ -23,15 +26,23 @@ enum State {
 /// ReplicaID is a type alias used to identify Raft nodes.
 pub type ReplicaID = usize;
 
+type Result<T> = std::result::Result<T, ReplicaError>;
+
+#[derive(Debug, Clone)]
+enum ReplicaError {
+    LogCompacted,
+}
+
 /// Replica describes the local instance running the Raft algorithm. Its goal is
 /// to maintain the consistency of the user-defined StateMachine across the
 /// cluster. It uses the user-defined Cluster implementation to talk to other
 /// Replicas, be it over the network or pigeon post.
-pub struct Replica<S, T, C>
+pub struct Replica<C, M, T, D>
 where
+    C: Cluster<T, D>,
+    M: StateMachine<T, D>,
     T: StateMachineTransition,
-    S: StateMachine<T>,
-    C: Cluster<T>,
+    D: Clone,
 {
     /// ID of this Replica.
     id: ReplicaID,
@@ -40,7 +51,7 @@ where
     peer_ids: Vec<ReplicaID>,
 
     /// User-defined state machine that the cluster Replicates.
-    state_machine: Arc<Mutex<S>>,
+    state_machine: Arc<Mutex<M>>,
 
     /// Interface a Replica uses to communicate with the rest of the cluster.
     cluster: Arc<Mutex<C>>,
@@ -89,13 +100,33 @@ where
     /// If no heartbeat message is received by the deadline, the Replica will
     /// start an election.
     next_election_deadline: Instant,
+
+    /// The number of transaction logs that this instance will let accumulate
+    /// before merging them into a single snapshot. Snapshotting is enabled <=>
+    /// snapshot_delta > 0.
+    snapshot_delta: usize,
+
+    /// The log snapshot of this Replica. Even if snapshot_delta is 0, the
+    /// snapshot field can be Some(_), since the Replica can be started with a
+    /// seed snapshot.
+    snapshot: Option<Snapshot<D>>,
+
+    /// The length of the log sequence that is represented by the snapshot.
+    /// Since compacted entries aren't in the log anymore, access to the log
+    /// should be done with log[log_index - index_offset].
+    ///
+    /// The following is always true:
+    ///
+    /// last_log_index = log.len() - 1 + index_offset.
+    index_offset: usize,
 }
 
-impl<S, T, C> Replica<S, T, C>
+impl<C, M, T, D> Replica<C, M, T, D>
 where
+    C: Cluster<T, D>,
+    M: StateMachine<T, D>,
     T: StateMachineTransition,
-    S: StateMachine<T>,
-    C: Cluster<T>,
+    D: Clone,
 {
     /// Create a new Replica.
     ///
@@ -108,6 +139,10 @@ where
     ///
     /// state_machine is the state machine that Raft maintains.
     ///
+    /// snapshot_delta tells the Replica how many transaction logs to accumulate
+    /// before doing compaction and merging them into a snapshot. Snapshotting
+    /// is enabled if and only if snapshot_delta > 0.
+    ///
     /// noop_transition is a transition that can be applied to the state machine
     /// multiple times with no effect.
     ///
@@ -116,37 +151,53 @@ where
     ///
     /// election_timeout_range defines the election timeout interval. If the
     /// Replica gets no messages from the Leader before the timeout, it
-    /// initiates an election.
-    ///
-    /// In practice, pick election_timeout_range to be 2-3x the value of
-    /// heartbeat_timeout, depending on your particular use-case network latency
-    /// and responsiveness needs. An election_timeout_range / heartbeat_timeout
-    /// ratio that's too low might cause unwarranted re-elections in the
-    /// cluster.
+    /// initiates an election. In practice, pick election_timeout_range to be
+    /// 2-3x the value of heartbeat_timeout, depending on your particular
+    /// use-case network latency and responsiveness needs. An
+    /// election_timeout_range / heartbeat_timeout ratio that's too low might
+    /// cause unwarranted re-elections in the cluster.
     pub fn new(
         id: ReplicaID,
         peer_ids: Vec<ReplicaID>,
         cluster: Arc<Mutex<C>>,
-        state_machine: Arc<Mutex<S>>,
+        state_machine: Arc<Mutex<M>>,
+        snapshot_delta: usize,
         noop_transition: T,
         heartbeat_timeout: Duration,
         election_timeout_range: (Duration, Duration),
-    ) -> Replica<S, T, C> {
-        Replica {
-            state_machine: state_machine,
-            cluster: cluster,
-            peer_ids: peer_ids,
-            id: id,
-            current_term: 0,
-            current_votes: None,
-            state: State::Follower,
-            voted_for: None,
-            log: vec![LogEntry {
+    ) -> Replica<C, M, T, D> {
+        let snapshot = state_machine.lock().unwrap().get_snapshot();
+        // index_offset is the "length" of the snapshot, so calculate it as
+        // snapshot.last_included_index + 1.
+        let mut index_offset: usize = 0;
+        let mut current_term: usize = 0;
+        let mut log: Vec<LogEntry<T>> = Vec::new();
+        if let Some(ref snapshot) = snapshot {
+            index_offset = snapshot.last_included_index + 1;
+            current_term = snapshot.last_included_term;
+        } else {
+            // If the Replica is starting anew, create a default no-op transition as
+            // the very first entry in the log. This trick lets us make sure every
+            // Replica has a non-empty log. If the Replica is starting from a
+            // snapshot, initialize current log to empty.
+            log = vec![LogEntry {
                 term: 0,
                 index: 0,
                 transition: noop_transition.clone(),
-            }],
-            noop_transition: noop_transition.clone(),
+            }]
+        }
+
+        Replica {
+            state_machine,
+            cluster,
+            peer_ids,
+            id,
+            current_term,
+            current_votes: None,
+            state: State::Follower,
+            voted_for: None,
+            log,
+            noop_transition,
             commit_index: 0,
             last_applied: 0,
             next_index: BTreeMap::new(),
@@ -154,6 +205,9 @@ where
             election_timeout: election_timeout_range,
             heartbeat_timer: Timer::new(heartbeat_timeout),
             next_election_deadline: Instant::now(),
+            snapshot,
+            snapshot_delta,
+            index_offset,
         }
     }
 
@@ -163,10 +217,6 @@ where
     /// new messages from the Cluster are available. The Replica will not poll
     /// for messages from the Cluster unless notified through recv_msg.
     ///
-    /// recv_msg is a channel on which the user must notify the Replica whenever
-    /// new messages from the Cluster are available. The Replica will not poll
-    /// for messages from the Cluster unless notified through recv_msg.
-    ///
     /// recv_transition is a channel on which the user must notify the Replica
     /// whenever new transitions to be processed for the StateMachine are
     /// available. The Replica will not poll for pending transitions for the
@@ -226,16 +276,46 @@ where
     }
 
     fn broadcast_append_entry_request(&mut self) {
-        self.broadcast_message(|peer_id: ReplicaID| Message::AppendEntryRequest {
-            term: self.current_term,
-            from_id: self.id,
-            prev_log_index: self.next_index[&peer_id] - 1,
-            prev_log_term: self.log[self.next_index[&peer_id] - 1].term,
-            entries: self.get_entries_for_peer(peer_id),
-            commit_index: self.commit_index,
+        self.broadcast_message(|peer_id: ReplicaID| {
+            match self.get_term_at_index(self.next_index[&peer_id] - 1) {
+                Ok(term) => Message::AppendEntryRequest {
+                    from_id: self.id,
+                    term: self.current_term,
+                    prev_log_index: self.next_index[&peer_id] - 1,
+                    prev_log_term: term,
+                    entries: self.get_entries_for_peer(peer_id),
+                    commit_index: self.commit_index,
+                },
+                Err(ReplicaError::LogCompacted) => {
+                    let snapshot = self.snapshot.as_ref().unwrap();
+                    Message::InstallSnapshotRequest {
+                        from_id: self.id,
+                        term: self.current_term,
+                        last_included_index: snapshot.last_included_index,
+                        last_included_term: snapshot.last_included_term,
+                        offset: 0,
+                        data: snapshot.data.clone(),
+                        done: true,
+                    }
+                }
+            }
         });
     }
 
+    fn get_term_at_index(&self, index: usize) -> Result<usize> {
+        if let Some(snapshot) = &self.snapshot {
+            if index == snapshot.last_included_index {
+                return Ok(snapshot.last_included_term);
+            } else if index > snapshot.last_included_index {
+                let localized_index = index - self.index_offset;
+                return Ok(self.log[localized_index].term);
+            }
+            Err(ReplicaError::LogCompacted)
+        } else {
+            Ok(self.log[index].term)
+        }
+    }
+
     fn poll_as_follower(&mut self, recv_msg: &Receiver<()>) {
         match recv_msg.recv_deadline(self.next_election_deadline) {
             // Process pending messages.
@@ -243,7 +323,7 @@ where
                 let messages = self.cluster.lock().unwrap().receive_messages();
                 // Update the election deadline if more than zero messages were
                 // actually received.
-                if messages.len() != 0 {
+                if !messages.is_empty() {
                     self.update_election_deadline();
                 }
 
@@ -264,7 +344,7 @@ where
         self.load_new_transitions();
     }
 
-    fn process_message(&mut self, message: Message<T>) {
+    fn process_message(&mut self, message: Message<T, D>) {
         match self.state {
             State::Leader => self.process_message_as_leader(message),
             State::Candidate => self.process_message_as_candidate(message),
@@ -285,7 +365,7 @@ where
                 let messages = self.cluster.lock().unwrap().receive_messages();
                 // Update the election deadline if more than zero messages were
                 // actually received.
-                if messages.len() != 0 {
+                if !messages.is_empty() {
                     self.update_election_deadline();
                 }
                 for message in messages {
@@ -307,27 +387,33 @@ where
 
     fn broadcast_message<F>(&self, message_generator: F)
     where
-        F: Fn(usize) -> Message<T>,
+        F: Fn(usize) -> Message<T, D>,
     {
         self.peer_ids.iter().for_each(|peer_id| {
             self.cluster
                 .lock()
                 .unwrap()
-                .send_message(peer_id.clone(), message_generator(peer_id.clone()))
+                .send_message(*peer_id, message_generator(*peer_id))
         });
     }
 
     // Get log entries that have not been acknowledged by the peer.
     fn get_entries_for_peer(&self, peer_id: ReplicaID) -> Vec<LogEntry<T>> {
-        self.log[self.next_index[&peer_id]..self.log.len()].to_vec()
+        // TODO: double check
+        self.log[self.next_index[&peer_id] - self.index_offset..self.log.len()].to_vec()
     }
 
     // Apply entries that are ready to be applied.
     fn apply_ready_entries(&mut self) {
+        if self.log.is_empty() {
+            return;
+        }
+
         // Move the commit index to the latest log index that has been
         // replicated on the majority of the replicas.
-        if self.state == State::Leader && self.commit_index < self.log.len() - 1 {
-            let mut n = self.log.len() - 1;
+        let mut state_machine = self.state_machine.lock().unwrap();
+        let mut n = self.log.len() - 1 + self.index_offset;
+        if self.state == State::Leader && self.commit_index < n {
             let old_commit_index = self.commit_index;
             while n > self.commit_index {
                 let num_replications =
@@ -337,7 +423,7 @@ where
                     );
 
                 if num_replications * 2 >= self.peer_ids.len()
-                    && self.log[n].term == self.current_term
+                    && self.log[n - self.index_offset].term == self.current_term
                 {
                     self.commit_index = n;
                 }
@@ -345,9 +431,8 @@ where
             }
 
             for i in old_commit_index + 1..=self.commit_index {
-                let mut state_machine = self.state_machine.lock().unwrap();
                 state_machine.register_transition_state(
-                    self.log[i].transition.get_id(),
+                    self.log[i - self.index_offset].transition.get_id(),
                     TransitionState::Committed,
                 );
             }
@@ -356,40 +441,63 @@ where
         // Apply entries that are behind the currently committed index.
         while self.commit_index > self.last_applied {
             self.last_applied += 1;
-            let mut state_machine = self.state_machine.lock().unwrap();
-            state_machine.apply_transition(self.log[self.last_applied].transition.clone());
+            let local_idx = self.last_applied - self.index_offset;
+            state_machine.apply_transition(self.log[local_idx].transition.clone());
             state_machine.register_transition_state(
-                self.log[self.last_applied].transition.get_id(),
+                self.log[local_idx].transition.get_id(),
                 TransitionState::Applied,
             );
         }
+
+        // If snapshot_delta is greater than 0, check whether it's time for log
+        // compaction.
+        if self.snapshot_delta > 0 {
+            // Calculate number of applied logs that haven't been compacted yet.
+            let curr_delta = self.last_applied + 1 - self.index_offset;
+            // If the number of accumulated logs is greater than or equal to the
+            // configured delta, do compaction.
+            if curr_delta >= self.snapshot_delta {
+                let last_applied = self.last_applied;
+                self.snapshot = Some(state_machine.create_snapshot(
+                    last_applied,
+                    self.log[last_applied - self.index_offset].term,
+                ));
+                self.log.retain(|l| l.index > last_applied);
+                self.index_offset = last_applied + 1;
+            }
+        }
     }
 
     fn load_new_transitions(&mut self) {
         // Load new transitions. Ignore the transitions if the replica is not
         // the Leader.
-        let transitions = self.state_machine.lock().unwrap().get_pending_transitions();
+        let mut state_machine = self.state_machine.lock().unwrap();
+        let transitions = state_machine.get_pending_transitions();
         for transition in transitions {
             if self.state == State::Leader {
                 self.log.push(LogEntry {
-                    index: self.log.len(),
+                    index: self.log.len() + self.index_offset,
                     transition: transition.clone(),
                     term: self.current_term,
                 });
 
-                let mut state_machine = self.state_machine.lock().unwrap();
                 state_machine
                     .register_transition_state(transition.get_id(), TransitionState::Queued);
+            } else {
+                state_machine.register_transition_state(
+                    transition.get_id(),
+                    TransitionState::Abandoned(TransitionAbandonedReason::NotLeader),
+                );
             }
         }
     }
 
-    fn process_message_as_leader(&mut self, message: Message<T>) {
+    fn process_message_as_leader(&mut self, message: Message<T, D>) {
         match message {
             Message::AppendEntryResponse {
                 from_id,
-                success,
                 term,
+                success,
                 last_index,
                 mismatch_index,
             } => {
@@ -405,19 +513,18 @@ where
                     // Update information about the peer's logs.
                     //
                     // If the mismatch_index is greater than or equal to the
-                    // existing next_index, then we know that this rejection is
-                    // a stray out-of-order or duplicate rejection, which we can
-                    // ignore. The reason we know that is because mismatch_index
-                    // is set by the follower to prev_log_index, which was in
-                    // turn set by the leader to next_index-1. Hence
-                    // mismatch_index can't be greater than or equal to
-                    // next_index.
+                    // existing next_index, then we know that this rejection is a
+                    // stray out-of-order or duplicate rejection, which we can
+                    // ignore. The reason we know that is because mismatch_index is
+                    // set by the follower to prev_log_index, which was in turn set
+                    // by the leader to next_index-1. Hence mismatch_index can't be
+                    // greater than or equal to next_index.
                     //
-                    // If the mismatch_index isn't stray, we set next_index to
-                    // the min of next_index and last_index; this is equivalent
-                    // to the Raft paper's guidance on decreasing next_index by
-                    // one at a time, but is more performant in cases when we
-                    // can cut straight to the follower's last_index+1.
+                    // If the mismatch_index isn't stray, we set next_index to the
+                    // min of next_index and last_index; this is equivalent to the
+                    // Raft paper's guidance on decreasing next_index by one at a
+                    // time, but is more performant in cases when we can cut
+                    // straight to the follower's last_index+1.
                     if let Some(mismatch_index) = mismatch_index {
                         if mismatch_index < self.next_index[&from_id] {
                             let next_index = cmp::min(mismatch_index, last_index + 1);
@@ -426,6 +533,20 @@ where
                     }
                 }
             }
+            Message::InstallSnapshotResponse {
+                from_id,
+                term,
+                last_included_index,
+            } => {
+                if term > self.current_term {
+                    // Become follower if another node's term is higher.
+                    self.cluster.lock().unwrap().register_leader(None);
+                    self.become_follower(term);
+                } else {
+                    self.next_index.insert(from_id, last_included_index + 1);
+                    self.match_index.insert(from_id, last_included_index);
+                }
+            }
             _ => {}
         }
     }
@@ -437,39 +558,9 @@ where
         last_log_index: usize,
         last_log_term: usize,
     ) {
-        if self.current_term > term {
-            // Do not vote for Replicas that are behind.
-            self.cluster.lock().unwrap().send_message(
-                from_id,
-                Message::VoteResponse {
-                    from_id: self.id,
-                    term: self.current_term,
-                    vote_granted: false,
-                },
-            );
-        } else if self.current_term < term {
-            // Become follower if the other replica's term is higher.
-            self.cluster.lock().unwrap().register_leader(None);
-            self.become_follower(term);
-        }
-
-        if self.voted_for == None || self.voted_for == Some(from_id) {
-            if self.log[self.log.len() - 1].index <= last_log_index
-                && self.log[self.log.len() - 1].term <= last_log_term
-            {
-                // If the criteria are met, grant the vote.
-                self.cluster.lock().unwrap().register_leader(None);
-                self.cluster.lock().unwrap().send_message(
-                    from_id,
-                    Message::VoteResponse {
-                        from_id: self.id,
-                        term: self.current_term,
-                        vote_granted: true,
-                    },
-                );
-                self.voted_for = Some(from_id);
-            } else {
-                // If the criteria are not met, do not grant the vote.
+        match self.current_term.cmp(&term) {
+            Ordering::Greater => {
+                // Do not vote for Replicas that are behind.
                 self.cluster.lock().unwrap().send_message(
                     from_id,
                     Message::VoteResponse {
@@ -479,17 +570,97 @@ where
                     },
                 );
             }
-        } else {
-            // If voted for someone else, don't grant the vote.
-            self.cluster.lock().unwrap().send_message(
+            Ordering::Less => {
+                // Become a follower if the other replica's term is higher.
+                self.cluster.lock().unwrap().register_leader(None);
+                self.become_follower(term);
+            }
+            _ => {}
+        }
+
+        let self_last_log_index = self.get_last_log_index();
+        let self_last_log_term = self.get_last_log_term();
+        if (self.voted_for == None || self.voted_for == Some(from_id))
+            && self_last_log_index <= last_log_index
+            && self_last_log_term <= last_log_term
+        {
+            // If the criteria are met, grant the vote.
+            let mut cluster = self.cluster.lock().unwrap();
+            cluster.register_leader(None);
+            cluster.send_message(
                 from_id,
                 Message::VoteResponse {
                     from_id: self.id,
                     term: self.current_term,
-                    vote_granted: false,
+                    vote_granted: true,
                 },
-            )
+            );
+            self.voted_for = Some(from_id);
+            return;
         }
+
+        // If the criteria are not met or if already voted for someone else, do
+        // not grant the vote.
+        self.cluster.lock().unwrap().send_message(
+            from_id,
+            Message::VoteResponse {
+                from_id: self.id,
+                term: self.current_term,
+                vote_granted: false,
+            },
+        );
+    }
+
+    fn process_install_snapshot_request_as_follower(
+        &mut self,
+        from_id: ReplicaID,
+        term: usize,
+        last_included_index: usize,
+        last_included_term: usize,
+        _offset: usize,
+        data: D,
+        _done: bool,
+    ) {
+        if self.current_term > term {
+            self.cluster.lock().unwrap().send_message(
+                from_id,
+                Message::InstallSnapshotResponse {
+                    from_id: self.id,
+                    term: self.current_term,
+                    last_included_index: self.get_last_log_index(),
+                },
+            );
+            return;
+        }
+
+        let snapshot = Snapshot {
+            last_included_index,
+            last_included_term,
+            data,
+        };
+
+        // Retain only logs not already in the snapshot. These logs are
+        // guaranteed to not be committed yet (otherwise we wouldn't be
+        // receiving the snapshot in the first place), so it is correct to
+        // restore StateMachine state from the snapshot.
+        let mut state_machine = self.state_machine.lock().unwrap();
+        self.log.retain(|l| l.index > last_included_index);
+        state_machine.set_snapshot(snapshot.clone());
+        self.snapshot = Some(snapshot);
+        self.index_offset = last_included_index + 1;
+        self.commit_index = last_included_index;
+        self.last_applied = last_included_index;
+        // It is likely that the snapshot contained new information, so we need
+        // to update our current term.
+        self.current_term = self.get_last_log_term();
+        self.cluster.lock().unwrap().send_message(
+            from_id,
+            Message::InstallSnapshotResponse {
+                from_id: self.id,
+                term: self.current_term,
+                last_included_index: self.get_last_log_index(),
+            },
+        );
     }
 
     fn process_append_entry_request_as_follower(
@@ -509,14 +680,17 @@ where
                     from_id: self.id,
                     term: self.current_term,
                     success: false,
-                    last_index: self.log.len() - 1,
+                    last_index: self.get_last_log_index(),
                     mismatch_index: None,
                 },
             );
             return;
+        }
+
         // If our log doesn't contain an entry at prev_log_index with the
         // prev_log_term term, reply false.
-        } else if prev_log_index >= self.log.len() || self.log[prev_log_index].term != prev_log_term
+        if prev_log_index >= self.log.len() + self.index_offset
+            || self.get_term_at_index(prev_log_index).unwrap() != prev_log_term
         {
             self.cluster.lock().unwrap().send_message(
                 from_id,
@@ -524,48 +698,54 @@ where
                     from_id: self.id,
                     term: self.current_term,
                     success: false,
-                    last_index: self.log.len() - 1,
+                    last_index: self.get_last_log_index(),
                     mismatch_index: Some(prev_log_index),
                 },
             );
             return;
         }
 
+        let mut state_machine = self.state_machine.lock().unwrap();
         for entry in entries {
             // Drop local inconsistent logs.
-            if entry.index < self.log.len() && entry.term != self.log[entry.index].term {
+            if entry.index <= self.get_last_log_index()
+            && entry.term != self.get_term_at_index(entry.index).unwrap() {
+                for i in entry.index..self.log.len() {
+                    state_machine.register_transition_state(
+                        self.log[i].transition.get_id(),
+                        TransitionState::Abandoned(TransitionAbandonedReason::ConflictWithLeader)
+                    );
+                } 
                 self.log.truncate(entry.index);
             }
 
             // Push received logs.
-            if entry.index == self.log.len() {
+            if entry.index == self.log.len() + self.index_offset {
                 self.log.push(entry);
             }
         }
 
         // Update local commit index to either the received commit index or the
         // latest local log position, whichever is smaller.
-        if commit_index > self.commit_index && self.log.len() != 0 {
-            self.commit_index = if commit_index < self.log[self.log.len() - 1].index {
-                commit_index
-            } else {
-                self.log[self.log.len() - 1].index
-            }
+        if commit_index > self.commit_index && !self.log.is_empty() {
+            self.commit_index = cmp::min(commit_index, self.log[self.log.len() - 1].index);
         }
-        self.cluster.lock().unwrap().register_leader(Some(from_id));
-        self.cluster.lock().unwrap().send_message(
+
+        let mut cluster = self.cluster.lock().unwrap();
+        cluster.register_leader(Some(from_id));
+        cluster.send_message(
             from_id,
             Message::AppendEntryResponse {
                 from_id: self.id,
                 term: self.current_term,
                 success: true,
-                last_index: self.log.len() - 1,
+                last_index: self.get_last_log_index(),
                 mismatch_index: None,
             },
         );
     }
 
-    fn process_message_as_follower(&mut self, message: Message<T>) {
+    fn process_message_as_follower(&mut self, message: Message<T, D>) {
         match message {
             Message::VoteRequest {
                 from_id,
@@ -590,12 +770,28 @@ where
                 entries,
                 commit_index,
             ),
-            Message::AppendEntryResponse { .. } => { /* ignore */ }
-            Message::VoteResponse { .. } => { /* ignore */ }
+            Message::InstallSnapshotRequest {
+                from_id,
+                term,
+                last_included_index,
+                last_included_term,
+                offset,
+                data,
+                done,
+            } => self.process_install_snapshot_request_as_follower(
+                from_id,
+                term,
+                last_included_index,
+                last_included_term,
+                offset,
+                data,
+                done,
+            ),
+            _ => { /* ignore */ }
         }
     }
 
-    fn process_message_as_candidate(&mut self, message: Message<T>) {
+    fn process_message_as_candidate(&mut self, message: Message<T, D>) {
         match message {
             Message::AppendEntryRequest { term, from_id, .. } => {
                 self.process_append_entry_request_as_candidate(term, from_id, message)
@@ -608,7 +804,35 @@ where
                 term,
                 vote_granted,
             } => self.process_vote_response_as_candidate(from_id, term, vote_granted),
-            Message::AppendEntryResponse { .. } => { /* ignore */ }
+            Message::InstallSnapshotRequest { from_id, term, .. } => {
+                self.process_install_snapshot_request_as_candidate(from_id, term, message)
+            }
+            _ => { /* ignore */ }
+        }
+    }
+
+    fn process_install_snapshot_request_as_candidate(
+        &mut self,
+        from_id: ReplicaID,
+        term: usize,
+        message: Message<T, D>,
+    ) {
+        // If the term is greater or equal to current term, then there's an
+        // active Leader, so convert self to a follower. If the term is smaller
+        // than the current term, inform the sender of your current term.
+        if term >= self.current_term {
+            self.cluster.lock().unwrap().register_leader(None);
+            self.become_follower(term);
+            self.process_message(message);
+        } else {
+            self.cluster.lock().unwrap().send_message(
+                from_id,
+                Message::InstallSnapshotResponse {
+                    from_id: self.id,
+                    last_included_index: self.get_last_log_index(),
+                    term: self.current_term,
+                },
+            );
         }
     }
 
@@ -639,7 +863,7 @@ where
         &mut self,
         term: usize,
         from_id: ReplicaID,
-        message: Message<T>,
+        message: Message<T, D>,
     ) {
         if term > self.current_term {
             self.cluster.lock().unwrap().register_leader(None);
@@ -661,7 +885,7 @@ where
         &mut self,
         term: usize,
         from_id: ReplicaID,
-        message: Message<T>,
+        message: Message<T, D>,
     ) {
         if term >= self.current_term {
             self.cluster.lock().unwrap().register_leader(None);
@@ -674,7 +898,7 @@ where
                     from_id: self.id,
                     term: self.current_term,
                     success: false,
-                    last_index: self.log.len() - 1,
+                    last_index: self.get_last_log_index(),
                     mismatch_index: None,
                 },
             );
@@ -689,8 +913,9 @@ where
         self.next_index = BTreeMap::new();
         self.match_index = BTreeMap::new();
         for peer_id in &self.peer_ids {
-            self.next_index.insert(peer_id.clone(), self.log.len());
-            self.match_index.insert(peer_id.clone(), 0);
+            self.next_index
+                .insert(*peer_id, self.log.len() + self.index_offset);
+            self.match_index.insert(*peer_id, 0);
         }
 
         // If the previous Leader had some uncommitted entries that were
@@ -700,7 +925,7 @@ where
         // emerges, append a no-op entry. This is a neat optimization described
         // in the part 8 of the paper.
         self.log.push(LogEntry {
-            index: self.log.len(),
+            index: self.log.len() + self.index_offset,
             transition: self.noop_transition.clone(),
             term: self.current_term,
         });
@@ -727,12 +952,28 @@ where
         self.broadcast_message(|_: usize| Message::VoteRequest {
             from_id: self.id,
             term: self.current_term,
-            last_log_index: self.log.len() - 1,
-            last_log_term: self.log[self.log.len() - 1].term,
+            last_log_index: self.get_last_log_index(),
+            last_log_term: self.get_last_log_term(),
         });
 
-        if self.peer_ids.len() == 0 {
+        if self.peer_ids.is_empty() {
             self.become_leader();
         }
     }
+
+    fn get_last_log_index(&self) -> usize {
+        if let Some(log) = self.log.last() {
+            log.index
+        } else {
+            self.index_offset - 1
+        }
+    }
+
+    fn get_last_log_term(&self) -> usize {
+        if let Some(log) = self.log.last() {
+            log.term
+        } else {
+            self.snapshot.as_ref().unwrap().last_included_term
+        }
+    }
 }
