@@ -1,16 +1,20 @@
+use bytes::Bytes;
 use crossbeam_channel as channel;
 use crossbeam_channel::{unbounded, Receiver, Sender};
+use rand::{thread_rng, Rng};
+use rand::seq::SliceRandom;
 use little_raft::{
     cluster::Cluster,
     message::Message,
     replica::Replica,
-    state_machine::{StateMachine, StateMachineTransition, TransitionState},
+    state_machine::{Snapshot, StateMachine, StateMachineTransition, TransitionState},
 };
+use std::convert::TryInto;
 use std::sync::{Arc, Mutex};
 
 use std::{collections::BTreeMap, thread, time::Duration};
 
-const HEARTBEAT_TIMEOUT: Duration = Duration::from_millis(500);
+const HEARTBEAT_TIMEOUT: Duration = Duration::from_millis(50);
 const MIN_ELECTION_TIMEOUT: Duration = Duration::from_millis(750);
 const MAX_ELECTION_TIMEOUT: Duration = Duration::from_millis(950);
 
@@ -38,9 +42,10 @@ struct Calculator {
     pending_transitions: Vec<ArithmeticOperation>,
 }
 
-impl StateMachine<ArithmeticOperation> for Calculator {
+impl StateMachine<ArithmeticOperation, Bytes> for Calculator {
     fn apply_transition(&mut self, transition: ArithmeticOperation) {
         self.value += transition.delta;
+        println!("id {} my value is now {} after applying delta {}", self.id, self.value, transition.delta);
     }
 
     fn register_transition_state(
@@ -62,18 +67,38 @@ impl StateMachine<ArithmeticOperation> for Calculator {
         self.pending_transitions = Vec::new();
         cur
     }
+
+    fn get_snapshot(&mut self) -> Option<Snapshot<Bytes>> {
+        println!("id {} checked for snapshot", self.id);
+        None
+    }
+
+    fn create_snapshot(&mut self, index: usize, term: usize) -> Snapshot<Bytes> {
+        println!("id {} created snapshot", self.id);
+        Snapshot {
+            last_included_index: index,
+            last_included_term: term,
+            data: Bytes::from(self.value.to_be_bytes().to_vec()),
+        }
+    }
+
+    fn set_snapshot(&mut self, snapshot: Snapshot<Bytes>) {
+        let v: Vec<u8> = snapshot.data.into_iter().collect();
+        self.value = i32::from_be_bytes(v[..].try_into().expect("incorrect length"));
+        println!("id {} my value is now {} after loading", self.id, self.value);
+    }
 }
 
 // Our test replicas will be running each in its own thread.
 struct ThreadCluster {
     id: usize,
     is_leader: bool,
-    transmitters: BTreeMap<usize, Sender<Message<ArithmeticOperation>>>,
-    pending_messages: Vec<Message<ArithmeticOperation>>,
+    transmitters: BTreeMap<usize, Sender<Message<ArithmeticOperation, Bytes>>>,
+    pending_messages: Vec<Message<ArithmeticOperation, Bytes>>,
     halt: bool,
 }
 
-impl Cluster<ArithmeticOperation> for ThreadCluster {
+impl Cluster<ArithmeticOperation, Bytes> for ThreadCluster {
     fn register_leader(&mut self, leader_id: Option<usize>) {
         if let Some(id) = leader_id {
             if id == self.id {
@@ -86,7 +111,13 @@ impl Cluster<ArithmeticOperation> for ThreadCluster {
         }
     }
 
-    fn send_message(&mut self, to_id: usize, message: Message<ArithmeticOperation>) {
+    fn send_message(&mut self, to_id: usize, message: Message<ArithmeticOperation, Bytes>) {
+        // Drop messages with probability 0.25.
+        let n: u8 = rand::thread_rng().gen();
+        if n % 4 == 0 {
+            return
+        }
+
         if let Some(transmitter) = self.transmitters.get(&to_id) {
             transmitter.send(message).expect("could not send message");
         }
@@ -96,8 +127,10 @@ impl Cluster<ArithmeticOperation> for ThreadCluster {
         self.halt
     }
 
-    fn receive_messages(&mut self) -> Vec<Message<ArithmeticOperation>> {
-        let cur = self.pending_messages.clone();
+    fn receive_messages(&mut self) -> Vec<Message<ArithmeticOperation, Bytes>> {
+        let mut cur = self.pending_messages.clone();
+        // Shuffle messages.
+        cur.shuffle(&mut thread_rng());
         self.pending_messages = Vec::new();
         cur
     }
@@ -107,7 +140,7 @@ impl Cluster<ArithmeticOperation> for ThreadCluster {
 // communication between replicas (threads).
 fn create_clusters(
     n: usize,
-    transmitters: BTreeMap<usize, Sender<Message<ArithmeticOperation>>>,
+    transmitters: BTreeMap<usize, Sender<Message<ArithmeticOperation, Bytes>>>,
 ) -> Vec<Arc<Mutex<ThreadCluster>>> {
     let mut clusters = Vec::new();
     for i in 0..n {
@@ -129,12 +162,12 @@ fn create_clusters(
 fn create_communication_between_clusters(
     n: usize,
 ) -> (
-    BTreeMap<usize, Sender<Message<ArithmeticOperation>>>,
-    Vec<Receiver<Message<ArithmeticOperation>>>,
+    BTreeMap<usize, Sender<Message<ArithmeticOperation, Bytes>>>,
+    Vec<Receiver<Message<ArithmeticOperation, Bytes>>>,
 ) {
     let (mut transmitters, mut receivers) = (BTreeMap::new(), Vec::new());
     for i in 0..n {
-        let (tx, rx) = unbounded::<Message<ArithmeticOperation>>();
+        let (tx, rx) = unbounded::<Message<ArithmeticOperation, Bytes>>();
         transmitters.insert(i, tx);
         receivers.push(rx);
     }
@@ -205,7 +238,7 @@ fn create_notifiers(
 
 fn run_clusters_communication(
     mut clusters: Vec<Arc<Mutex<ThreadCluster>>>,
-    mut cluster_message_receivers: Vec<Receiver<Message<ArithmeticOperation>>>,
+    mut cluster_message_receivers: Vec<Receiver<Message<ArithmeticOperation, Bytes>>>,
     mut message_notifiers_tx: Vec<Sender<()>>,
 ) {
     for _ in (0..clusters.len()).rev() {
@@ -237,7 +270,8 @@ fn run_arithmetic_operation_on_cluster(
     delta: i32,
     id: usize,
 ) {
-    thread::sleep(Duration::from_secs(1));
+    // Sleep longer because in this test we're dropping 25% of all messages.
+    thread::sleep(Duration::from_secs(2));
     // Find the leader and send the transition request to it.
     for cluster in clusters.iter() {
         let cluster = cluster.lock().unwrap();
@@ -246,10 +280,7 @@ fn run_arithmetic_operation_on_cluster(
                 .lock()
                 .unwrap()
                 .pending_transitions
-                .push(ArithmeticOperation {
-                    delta: delta,
-                    id: id,
-                });
+                .push(ArithmeticOperation { delta, id });
             transition_notifiers[cluster.id]
                 .send(())
                 .expect("could not send transition notification");
@@ -257,7 +288,8 @@ fn run_arithmetic_operation_on_cluster(
         }
     }
 
-    thread::sleep(Duration::from_secs(2));
+    // Sleep long.
+    thread::sleep(Duration::from_secs(3));
 }
 
 fn halt_clusters(clusters: Vec<Arc<Mutex<ThreadCluster>>>) {
@@ -284,7 +316,7 @@ fn run_replicas() {
     let clusters = create_clusters(n, transmitters);
     let peer_ids = create_peer_ids(n);
     let noop = ArithmeticOperation { delta: 0, id: 0 };
-    let (applied_transitions_tx, applied_transitions_rx) = unbounded();
+    let (applied_transitions_tx, _applied_transitions_rx) = unbounded();
     let state_machines = create_state_machines(n, applied_transitions_tx);
     let (message_tx, transition_tx, message_rx, transition_rx) = create_notifiers(n);
     for i in 0..n {
@@ -300,6 +332,7 @@ fn run_replicas() {
                 local_peer_ids,
                 cluster,
                 state_machine,
+                1,
                 noop.clone(),
                 HEARTBEAT_TIMEOUT,
                 (MIN_ELECTION_TIMEOUT, MAX_ELECTION_TIMEOUT),
@@ -309,8 +342,7 @@ fn run_replicas() {
         });
     }
 
-    run_clusters_communication(clusters.clone(), receivers.clone(), message_tx);
-
+    run_clusters_communication(clusters.clone(), receivers, message_tx);
     run_arithmetic_operation_on_cluster(
         clusters.clone(),
         state_machines.clone(),
@@ -319,12 +351,16 @@ fn run_replicas() {
         1,
     );
 
-    // Signal to the 0th replica that it should halt, give the remaining
-    // replicas some time to reelect the leader, and mark the 0th replica as a
-    // non-leader.
-    clusters.clone()[0].lock().unwrap().halt = true;
-    thread::sleep(Duration::from_secs(2));
-    clusters.clone()[0].lock().unwrap().is_leader = false;
+    // In this test, we confirm that the cluster converged on true value one by
+    // one after each arithmetic operation. This is different from
+    // raft_stable.rs, where we check the order in which transition have been
+    // applied post-factum. We can't do the same in raft_unstable.rs, because
+    // replicas reload from snapshots in this test, meaning not all replicas go
+    // over all transitions. Some replicas load directly from their peer's
+    // snapshots.
+    for machine in state_machines.clone() {
+        assert_eq!(machine.lock().unwrap().value, 5);
+    }
 
     run_arithmetic_operation_on_cluster(
         clusters.clone(),
@@ -334,6 +370,10 @@ fn run_replicas() {
         2,
     );
 
+    for machine in state_machines.clone() {
+        assert_eq!(machine.lock().unwrap().value, -46);
+    }
+
     run_arithmetic_operation_on_cluster(
         clusters.clone(),
         state_machines.clone(),
@@ -342,37 +382,16 @@ fn run_replicas() {
         3,
     );
 
-    run_arithmetic_operation_on_cluster(
-        clusters.clone(),
-        state_machines.clone(),
-        transition_tx.clone(),
-        3,
-        4,
-    );
 
-    halt_clusters(clusters.clone());
+    for machine in state_machines.clone() {
+        assert_eq!(machine.lock().unwrap().value, -557);
+    }
 
-    // Below we confirm that every replica applied the same transitions in the
-    // same order.
-    let applied_transactions: Vec<(usize, usize)> = applied_transitions_rx.try_iter().collect();
-    let expected_vec: Vec<usize> = vec![1, 2, 3, 4];
-    assert_eq!(
-        expected_vec,
-        applied_transactions.iter().fold(Vec::new(), |mut acc, x| {
-            if x.0 == 1 && x.1 != 0 {
-                acc.push(x.1);
-            };
-            acc
-        })
-    );
+    run_arithmetic_operation_on_cluster(clusters.clone(), state_machines.clone(), transition_tx.clone(), 3, 4);
 
-    assert_eq!(
-        expected_vec,
-        applied_transactions.iter().fold(Vec::new(), |mut acc, x| {
-            if x.0 == 2 && x.1 != 0 {
-                acc.push(x.1);
-            };
-            acc
-        })
-    );
+    for machine in state_machines.clone() {
+        assert_eq!(machine.lock().unwrap().value, -554);
+    }
+
+    halt_clusters(clusters);
 }
